using System;
using System.Diagnostics;
using System.IO;
using System.ServiceModel;
using System.Text;
using System.Xml;
using Microsoft.BizTalk.Component.Interop;
using Microsoft.BizTalk.Message.Interop;
using Microsoft.BizTalk.Streaming;
using Microsoft.BizTalk.TransportProxy.Interop;
using Microsoft.Samples.BizTalk.Adapter.Common;
using Microsoft.Samples.BizTalk.Adapters.Wcf.Configuration;
using Microsoft.Samples.BizTalk.Adapters.Wcf.Utility;
using System.ServiceModel.Channels;
using System.ServiceModel.Dispatcher;

namespace Microsoft.Samples.BizTalk.Adapters.Wcf.Receiver
{
    internal sealed class WcfReceiverEndpoint : ReceiverEndpoint
    {
        private const string MESSAGE_BODY = "body";

        private IBTTransportProxy _transportProxy;
        private WcfReceiverProperties _properties;
        private ControlledTermination _control;
        private string _propertyNamespace;
        private string _uri;
        private IBaseMessageFactory _messageFactory;
        private string _transportType;



        #region Open / Update / Dispose

        public override void Open(string uri, IPropertyBag config, IPropertyBag bizTalkConfig,
                                  IPropertyBag handlerPropertyBag, IBTTransportProxy transportProxy,
                                  string transportType, string propertyNamespace, ControlledTermination control)
        {
            _properties = new WcfReceiverProperties(uri);

            //  Handler properties
            //            XmlDocument handlerConfigDom = ConfigProperties.IfExistsExtractConfigDom(handlerPropertyBag);
            //            if (null != handlerConfigDom)
            //                _properties.HandlerConfiguration(handlerConfigDom);

            //  Location properties - possibly override some Handler properties
            //            XmlDocument locationConfigDom = ConfigProperties.ExtractConfigDom(config);
            _properties.LocationConfiguration(config, bizTalkConfig);

            _transportProxy = transportProxy;
            _control = control;

            _uri = uri;
            _transportType = transportType;
            _propertyNamespace = propertyNamespace;
            _messageFactory = _transportProxy.GetMessageFactory();

            Start();
        }

        public override void Update(IPropertyBag config, IPropertyBag bizTalkConfig, IPropertyBag handlerPropertyBag)
        {
            Stop();

            _properties = new WcfReceiverProperties(_properties.Uri);

            //  Handler properties
            //            XmlDocument handlerConfigDom = ConfigProperties.IfExistsExtractConfigDom(handlerPropertyBag);
            //            if (null != handlerConfigDom)
            //                _properties.HandlerConfiguration(handlerConfigDom);

            //  Location properties - possibly override some Handler properties
            _properties.LocationConfiguration(config, bizTalkConfig);

            Start();
        }

        public override void Dispose()
        {
            Stop();
            base.Dispose();
        }

        #endregion // Open / Update / Dispose

        private ServiceHost _host;

        private void Start()
        {
            Tracer.TraceMessage(TraceLevel.Info, "WCF: Starting host");

            // Starting i singleton RR receive service
            ServiceConfiguration serviceConfiguration = ServiceConfiguration.CreateFromProperties(_properties);

            WcfServiceFactory factory = new WcfServiceFactory(serviceConfiguration);
            if (_properties.IsTwoWay)
            {
                Tracer.TraceMessage("WCF: Setting up request/response service it {0}", serviceConfiguration.EndpointUri);

                IndigoRequestResponseService rrService = new IndigoRequestResponseService(this);
                _host =
                    factory.CreateServiceHost<IndigoRequestResponseService, IIndigoRequestResponseService>(rrService);
            }
            else
            {
                Tracer.TraceMessage("WCF: Setting up receive service it {0}", serviceConfiguration.EndpointUri);

                if (serviceConfiguration.UseNativeOneWay)
                {
                    IndigoReceiveServiceOneWay receiveService = new IndigoReceiveServiceOneWay(this);
                    _host = factory.CreateServiceHost<IndigoReceiveServiceOneWay, IIndigoReceiveServiceOneWay>(
                        receiveService);
                }
                else
                {
                    IndigoReceiveService receiveService = new IndigoReceiveService(this);
                    _host = factory.CreateServiceHost<IndigoReceiveService, IIndigoReceiveService>(receiveService);
                }
            }

            _host.Open();
            TraceEndpoints(_host.ChannelDispatchers);
        }

        private void Stop()
        {
            Tracer.TraceMessage(TraceLevel.Info, "WCF: Shutting down receive service.");

            if (_host != null)
            {
                _host.Close();
                ((IDisposable)_host).Dispose();
            }
        }

        #region Utility

        /// <summary>
        /// Creates a BizTalk message and populates the message context _properties
        /// </summary>
        /// <returns>BizTalk message</returns>
        private IBaseMessage CreateMessage(Stream msgData)
        {
            IBaseMessagePart part = _messageFactory.CreateMessagePart();
            part.Data = msgData;

            IBaseMessage message = _messageFactory.CreateMessage();
            message.AddPart(MESSAGE_BODY, part, true);

            SystemMessageContext context = new SystemMessageContext(message.Context);
            context.InboundTransportLocation = _uri;
            context.InboundTransportType = _transportType;

            return message;
        }


        private static void StreamCopy(Stream dest, Stream src)
        {
            const int DATA_BLOCK = 4096;
            int bytesRead;
            byte[] buff = new byte[DATA_BLOCK];
            while ((bytesRead = src.Read(buff, 0, DATA_BLOCK)) > 0)
                dest.Write(buff, 0, bytesRead);
        }

        /// <summary>
        /// Copies an Indigo message into a VirtualStream object.
        /// </summary>
        private static VirtualStream CopyMessageToVirtualStream(Message inputMessage)
        {
            XmlReader reader = inputMessage.GetReaderAtBodyContents();

            // Copy message to virtual stream
            VirtualStream newMessageStream = new VirtualStream();
            XmlTextWriter xmlTextWriter = new XmlTextWriter(newMessageStream, Encoding.UTF8);
            xmlTextWriter.WriteNode(reader, false);
            xmlTextWriter.Flush();

            newMessageStream.Flush();
            newMessageStream.Seek(0, SeekOrigin.Begin);

            return newMessageStream;
        }

        private static void TraceEndpoints(ChannelDispatcherCollection listeners)
        {
            foreach (ChannelDispatcher r in listeners)
            {
                foreach (EndpointDispatcher d in r.Endpoints)
                {
                    Tracer.TraceMessage(TraceLevel.Info, "Running service at {0}", d.EndpointAddress.Uri);
                }
            }
        }

        #endregion

        private object syncLock = new object();

        public Message ProcessRequestResponseMessage(Message requestMessage)
        {
            Tracer.TraceMessage(TraceLevel.Info, "WCF Listener Service: RR Message Received");

            try
            {
                // Copy message to virtual stream
                Stream requestMessageStream = CopyMessageToVirtualStream(requestMessage);
                IBaseMessage btMessage = CreateMessage(requestMessageStream);


                //Submit the message and wait for the answer
                //This has to be synchonized because the baseclasses
                //do not accept more than one SubmitRequestMessage/GetResponseMessage at a time.
                //Using an isolated adapter for request/response is a much better scenario.
                lock (syncLock)
                {
                    WcfRequestResponseHandler rrHandler = new WcfRequestResponseHandler(_transportProxy, _control);
                    rrHandler.SubmitRequestMessage(btMessage, null);


                    // Note: GetResponseMessage() is a blocking call
                    IBaseMessage responseBtsMessage = rrHandler.GetResponseMessage();
                    Debug.WriteLine("WCF Listener Service: BizTalk - GetResponseMessage");

                    if (responseBtsMessage == null)
                    {
                        Tracer.TraceMessage(
                            "WCFReceiver.ProcessRequestResponse(): failed to process the Indigo Request!",
                            "WCFReceive: Error");
                        throw new Exception("Failed to process the WCF Request");
                    }

                    // Copy the data from the response message into the response stream. 
                    // Note, the data is handled in a streaming fashion to 
                    // prevent us going out of memory for large messages.
                    Stream responseBtsMessageStream = responseBtsMessage.BodyPart.GetOriginalDataStream();

                    Stream responseStream = new VirtualStream();
                    StreamCopy(responseStream, responseBtsMessageStream);
                    responseStream.Flush();
                    responseStream.Seek(0, SeekOrigin.Begin);

                    rrHandler.ResponseProcessed();

                    Message replyMessage =
                        Message.CreateMessage(MessageVersion.Default, WcfReceiverSettings.ReplyAction, new XmlTextReader(responseStream));

                    Tracer.TraceMessage(TraceLevel.Info, "WCF Listener Service: Return Reply Message");

                    return replyMessage;
                }
            }
            catch (Exception ex)
            {
                Tracer.TraceMessage(TraceLevel.Error, "WCF Listener Service: {0}", ex);

                EventLog.WriteEntry("WCF", "GetReqRspMessagesAndSubmit Error: " + ex.Message, EventLogEntryType.Error);
                throw;
            }
        }


        /// <summary>
        /// Submit a single message to BTS
        /// </summary>
        /// <param name="requestMessage"></param>
        public void ProcessRequestMessage(Message requestMessage)
        {
            Stream requestMessageStream = CopyMessageToVirtualStream(requestMessage);

            ReceiveBatch batchHandler =
                new ReceiveBatch(_transportProxy, _control, (ReceiveBatchCompleteHandler) null, 1);

            try
            {
                IBaseMessage btMessage = CreateMessage(requestMessageStream);
                batchHandler.SubmitMessage(btMessage, null);
            }
            catch (Exception e)
            {
                Tracer.TraceMessage(TraceLevel.Error, e);
                EventLog.WriteEntry("IndigoAdapter", String.Format("Could not read message!"), EventLogEntryType.Error);
            }

            batchHandler.Done(null);
        }
    }
}