using System;
using System.Diagnostics;
using System.IO;
using System.ServiceModel.Channels;
using System.Text;
using System.Xml;
using Microsoft.BizTalk.Component.Interop;
using Microsoft.BizTalk.Message.Interop;
using Microsoft.BizTalk.Streaming;
using Microsoft.Samples.BizTalk.Adapter.Common;
using Microsoft.Samples.BizTalk.Adapters.Wcf.Configuration;

namespace Microsoft.Samples.BizTalk.Adapters.Wcf.Transmitter
{
    internal class WcfTransmitterEndpoint : AsyncTransmitterEndpoint
    {
        private AsyncTransmitter _asyncTransmitter;
        private string _propertyNamespace;

        public WcfTransmitterEndpoint(AsyncTransmitter asyncTransmitter)
            : base(asyncTransmitter)
        {
            _asyncTransmitter = asyncTransmitter;
        }

        public override void Open(EndpointParameters endpointParameters, IPropertyBag handlerPropertyBag,
                                  string propertyNamespace)
        {
            _propertyNamespace = propertyNamespace;
        }

        public override IBaseMessage ProcessMessage(IBaseMessage message)
        {
            WcfTransmitterProperties props = new WcfTransmitterProperties(message, _propertyNamespace);

            if (props.IsTwoWay)
            {
                return SendRequestResponse(message, props);
            }
            else
            {
                SendRequest(message, props);
                return null;
            }
        }

        private IBaseMessage SendRequestResponse(IBaseMessage msg, WcfTransmitterProperties config)
        {
            IBaseMessage response;
            if (!SendSolicitResponse(msg, out response, config))
                throw new ApplicationException("WCF: Failed to process");
            return response;
        }

        private void SendRequest(IBaseMessage msg, WcfTransmitterProperties config)
        {
            if (!SendOneWay(msg, config))
                throw new ApplicationException("WCF: Failed to process");
        }

        #region Send Messages

        /// <summary>
        /// Used to transmit an individual message. The message stream is read into 
        /// a buffer and written to disc. Note, we need to process the data in a 
        /// streaming fashion, since the stream we receive is most likely a forward- 
        /// only stream we have no idea how big it is. We therefore need to process
        /// it in a streaming fashion to avoid an out-of-memory (OOM) situation
        /// </summary>
        /// <param name="msg"></param>
        /// <returns></returns>
        private bool SendOneWay(IBaseMessage msg, WcfTransmitterProperties properties)
        {
            Stream responseStream;
            return SendIndigoRequest(msg, false, out responseStream, properties);
        }

        private bool SendSolicitResponse(IBaseMessage msg, out IBaseMessage responseMsg,
                                         WcfTransmitterProperties properties)
        {
            responseMsg = null;
            Stream responseStream;

            if (SendIndigoRequest(msg, true, out responseStream, properties))
            {
                IBaseMessageFactory mf = _asyncTransmitter.TransportProxy.GetMessageFactory();
                IBaseMessage newResponseMsg = mf.CreateMessage();
                IBaseMessagePart body = mf.CreateMessagePart();
                body.Data = responseStream;
                newResponseMsg.AddPart("Body", body, true);
                responseMsg = newResponseMsg;
                return true;
            }
            else
            {
                return false;
            }
        }

        private static bool SendIndigoRequest(IBaseMessage msg, bool requestResponseExchange, out Stream responseStream,
                                       WcfTransmitterProperties config)
        {
            Trace.WriteLine("Indigo Transmit Adapter: Send Indigo Request");
            responseStream = null;

            IBaseMessagePart bodyPart = msg.BodyPart;
            if (bodyPart == null)
                return false;

            Stream stream = bodyPart.GetOriginalDataStream();
            if (stream == null)
                return false;

            ServiceConfiguration serviceConfiguration = ServiceConfiguration.CreateFromProperties(config);
            WcfTransmitClient serviceProxy = new WcfTransmitClient(serviceConfiguration);

            if (requestResponseExchange)
            {
                Message responseMsg;
                using (XmlTextReader xmlReader = new XmlTextReader(stream))
                {
                    responseMsg = serviceProxy.SendRequestResponseMessage(xmlReader);
                }

                responseStream = new VirtualStream();

                XmlTextWriter writer = new XmlTextWriter(responseStream, Encoding.UTF8);
                writer.WriteNode(responseMsg.GetReaderAtBodyContents(), false);
                writer.Flush();

                responseStream.Seek(0, SeekOrigin.Begin);
            }
            else
            {
                using (XmlTextReader xmlReader = new XmlTextReader(stream))
                {
                    serviceProxy.SendOneWayMessage(xmlReader);
                }
            }

            return true;
        }

        #endregion // Send Messages
    }
}