﻿#region Using directives

using System;
using System.ServiceModel;
using Microsoft.Samples.BizTalk.Adapters.Wcf.Configuration;
using System.ServiceModel.Channels;
using System.ServiceModel.Description;

#endregion

namespace Microsoft.Samples.BizTalk.Adapters.Wcf.Utility
{
    internal sealed class WcfServiceFactory
    {
        private readonly ServiceConfiguration _serviceConfiguration;

        public WcfServiceFactory(ServiceConfiguration serviceConfiguration)
        {
            _serviceConfiguration = serviceConfiguration;
        }

        private ServiceConfiguration Configuration
        {
            get { return _serviceConfiguration; }
        }

        private Binding CreateBinding()
        {
            Binding binding;

            if (String.IsNullOrEmpty(Configuration.BindingConfiguration))
            {
                binding = Activator.CreateInstance(Configuration.BindingType) as Binding;
            }
            else
            {
                binding =
                    Activator.CreateInstance(Configuration.BindingType, Configuration.BindingConfiguration) as Binding;
            }

            if (binding == null)
                throw new ApplicationException("Unable to create binding.");

            return binding;
        }


        private EndpointIdentity CreateIdentity()
        {
            switch (Configuration.IdentityType)
            {
                case IdentityType.None:
                    return null;
                case IdentityType.Dns:
                    return EndpointIdentity.CreateDnsIdentity(Configuration.IdentityValue);
                case IdentityType.Rsa:
                    return EndpointIdentity.CreateRsaIdentity(Configuration.IdentityValue);
                case IdentityType.Spn:
                    return EndpointIdentity.CreateSpnIdentity(Configuration.IdentityValue);
                case IdentityType.Upn:
                    return EndpointIdentity.CreateUpnIdentity(Configuration.IdentityValue);
//                case IdentityType.X509Certificate:
//                    return Identity.CreateX509CertificateIdentity(Configuration.IdentityValue);
                default:
                    throw
                        new NotSupportedException(String.Format("Invalid Identity Type {0}", Configuration.IdentityType)
                            );
            }
        }

        public ServiceHost CreateServiceHost<ServiceType, ContractType>(ServiceType instance)
        {
            ServiceHost host = new ServiceHost(instance);
            ServiceEndpoint endpoint = host.AddServiceEndpoint(typeof(ContractType), CreateBinding(), Configuration.EndpointUri);

            EndpointIdentity identity = CreateIdentity();
            if (identity != null)
            {
                EndpointAddressBuilder addressBuilder = new EndpointAddressBuilder(endpoint.Address);
                addressBuilder.Identity = identity;
                endpoint.Address = addressBuilder.ToEndpointAddress();
            }

            if (!String.IsNullOrEmpty(Configuration.BehaviorConfiguration))
            {
                BehaviorManager manager = new BehaviorManager(Configuration.BehaviorConfiguration);
                manager.FixBehavior(host.Description);
            }

            Tracer.TraceMessage("Service configured at {0}", Configuration.EndpointUri);

            return host;
        }

        public ClientContractType CreateClient<ClientContractType>()
        {
            EndpointAddress address;
            EndpointIdentity identity = CreateIdentity();
            if (identity == null)
            {
                address = new EndpointAddress(Configuration.EndpointUri);
            }
            else
            {
                address = new EndpointAddress(Configuration.EndpointUri, identity);
            }

            ChannelFactory<ClientContractType> channelFactory =
                new ChannelFactory<ClientContractType>(CreateBinding(), address);
            if (!String.IsNullOrEmpty(Configuration.BehaviorConfiguration))
            {
                BehaviorManager manager = new BehaviorManager(Configuration.BehaviorConfiguration);
                manager.FixBehavior(channelFactory.Endpoint);
            }

            return channelFactory.CreateChannel();
        }
    }
}